import torch
import torch.nn as nn
from torch.nn import functional as F
import torchvision.transforms._functional_tensor as F_t

from typing import Any, Dict, Optional
from kornia.core import Tensor
import random
import math

from auglab.transforms.gpu.base import ImageOnlyTransform
from typing import Any, Dict, Optional, Tuple, Union, List

def _choose_region_mode(p_in: float, p_out: float, seg_mask: Optional[torch.Tensor]) -> str:
    """Sample where to apply the transform: 'in', 'out', or 'all'.

    - p_in, p_out are probabilities in [0,1].
    - If seg_mask is None, or both probs are 0, return 'all'.
    - If p_in + p_out > 1, renormalize so p_all=0.
    """
    p_in = float(max(0.0, min(1.0, p_in)))
    p_out = float(max(0.0, min(1.0, p_out)))
    in_bool = torch.rand(()) < p_in
    out_bool = torch.rand(()) < p_out
    if in_bool and not out_bool:
        return 'in'
    if out_bool and not in_bool:
        return 'out'
    return 'all'

def _apply_region_mode(orig: torch.Tensor, transformed: torch.Tensor, seg_mask: Optional[torch.Tensor], mode: str, normalize: bool = True, mix_in_out: bool = False) -> torch.Tensor:
    """Blend transformed with orig based on region selection mode.

    - mode 'all': return transformed
    - mode 'in': apply transform inside seg, keep orig outside
    - mode 'out': apply  transform outside seg, keep orig inside

    mix_in_out: if True, randomly apply transform to some of the segmentation, not all.
    """
    if seg_mask is None or mode == 'all':
        return transformed
    
    # Rescale transformed based on min max orig
    # Needed due to the important change in the image
    if orig.dim() == 4:
        if normalize:
            orig_min = torch.amin(orig, dim=tuple(range(1, orig.dim())), keepdim=True)
            orig_max = torch.amax(orig, dim=tuple(range(1, orig.dim())), keepdim=True)
            transformed_min = torch.amin(transformed, dim=tuple(range(1, transformed.dim())), keepdim=True)
            transformed_max = torch.amax(transformed, dim=tuple(range(1, transformed.dim())), keepdim=True)
            transformed = (transformed - transformed_min) / (transformed_max - transformed_min + 1e-8) * (orig_max - orig_min) + orig_min

        m = seg_mask.to(transformed.dtype)
        if mode == 'out':
            m = 1.0 - m
        if mix_in_out:
            for i in range(seg_mask.shape[0]):
                # Create a tensor with random one and zero
                
                o = torch.randint(0, 2, (seg_mask.shape[1],), device=seg_mask.device, dtype=seg_mask.dtype)
                m[i] = m[i] * o.view(-1, 1, 1, 1)  # Broadcasting o to match the dimensions of m
        
        m = torch.sum(m, axis=1)
    elif orig.dim() == 3:
        if normalize:
            orig_min = torch.amin(orig)
            orig_max = torch.amax(orig)
            transformed_min = torch.amin(transformed)
            transformed_max = torch.amax(transformed)
            transformed = (transformed - transformed_min) / (transformed_max - transformed_min + 1e-8) * (orig_max - orig_min) + orig_min
        
        m = seg_mask.to(transformed.dtype)
        if mode == 'out':
            m = 1.0 - m
        if mix_in_out:
            # Create a tensor with random one and zero
            o = torch.randint(0, 2, (seg_mask.shape[0],), device=seg_mask.device, dtype=seg_mask.dtype)
            m = m * o.view(-1, 1, 1, 1)  # Broadcasting o to match the dimensions of m
        m = torch.sum(m, axis=0)
    else:
        raise ValueError(f"Only 4D and 3D images are supported. Got {orig.dim()}D.")

    return m * transformed + (1.0 - m) * orig

## Convolution transform
class RandomConvTransformGPU(ImageOnlyTransform):
    """Apply convolution to image.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.
    Based on https://docs.pytorch.org/vision/0.9/transforms.html#torchvision.transforms.GaussianBlur

    Args:
        kernel_type (str): Type of convolution kernel, either 'Laplace' or 'Scharr'. Default is 'Laplace'.
        spatial_dims (int): Number of spatial dimensions of the input image, either 2 or 3. Default is 2.
        absolute (bool): If True, take the absolute value of the convolution result. Default is False.
        retain_stats (bool): If True, retain the original mean and standard deviation of the image after convolution. Default is False.

    Returns:
        Tensor: Convolved version of the input image.

    """
    def __init__(
        self,
        kernel_type: str = 'Laplace', 
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        same_on_batch: bool = False,
        retain_stats: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        if kernel_type not in ["Laplace", "Scharr", "GaussianBlur", "UnsharpMask", "RandConv"]:
            raise NotImplementedError('Currently only "Laplace", "Scharr", "GaussianBlur", "UnsharpMask" and "RandConv" are supported.')
        else:
            self.kernel_type = kernel_type
        self.apply_to_channel = apply_to_channel
        self.absolute = kwargs.get('absolute', False)
        self.sigma = kwargs.get('sigma', 1.0)
        self.retain_stats = retain_stats
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out
        # Unsharp mask parameters: amount controls strength of the mask
        self.unsharp_amount = kwargs.get('unsharp_amount', 1.0)
        # RandConv parameters
        self.kernel_sizes = kwargs.get("kernel_sizes", [1,3,5,7])  # multi-scale default
        self.mix_prob = kwargs.get("mix_prob", 0.0)  # probability to mix with original

    def get_kernel(self, device: torch.device) -> torch.Tensor:
        if self.kernel_type == "Laplace":
            kernel = -1.0 * torch.ones(3, 3, 3, dtype=torch.float32, device=device)
            kernel[1, 1, 1] = 26.0
        elif self.kernel_type == "Scharr":
            kernel_x = torch.tensor([[[  9,    0,    -9],
                                        [ 30,    0,   -30],
                                        [  9,    0,    -9]],

                                        [[ 30,    0,   -30],
                                        [100,    0,  -100],
                                        [ 30,    0,   -30]],

                                        [[  9,    0,    -9],
                                        [ 30,    0,   -30],
                                        [  9,    0,    -9]]], dtype=torch.float32, device=device)
            
            kernel_y = torch.tensor([[[    9,   30,    9],
                                        [    0,    0,    0],
                                        [   -9,  -30,   -9]],

                                        [[  30,  100,   30],
                                        [   0,    0,    0],
                                        [ -30, -100,  -30]],

                                        [[   9,   30,    9],
                                        [   0,    0,    0],
                                        [  -9,  -30,   -9]]], dtype=torch.float32, device=device)

            kernel_z = torch.tensor([[[   9,   30,   9],
                                        [  30,  100,  30],
                                        [   9,   30,   9]],

                                        [[   0,    0,   0],
                                        [   0,    0,   0],
                                        [   0,    0,   0]],

                                        [[   -9,  -30,  -9],
                                        [  -30, -100, -30],
                                        [   -9,  -30,  -9]]], dtype=torch.float32, device=device)
            kernel = [kernel_x, kernel_y, kernel_z]
        elif self.kernel_type == "GaussianBlur":
            sigma = torch.rand(3, device=device) * self.sigma
            kernel_size = 3
            kernel = get_gaussian_kernel3d(kernel_size, sigma, torch.float32, device)
        elif self.kernel_type == "UnsharpMask":
            # For unsharp masking we use a Gaussian blur kernel; amount is applied in apply_transform.
            sigma = torch.rand(3, device=device) * self.sigma
            kernel_size = 3
            kernel = get_gaussian_kernel3d(kernel_size, sigma, torch.float32, device)
        elif self.kernel_type == "RandConv":
            # choose random odd kernel size e.g. [1,3,5,7]
            k = int(random.choice(self.kernel_sizes))  # define kernel_sizes in __init__

            std = 1.0 / math.sqrt(k * k)
            kernel = torch.randn(
                (k, k, k),  # for 3D
                device=device
            ) * std
        else:
            raise NotImplementedError('Kernel type not implemented.')
        return kernel
    
    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:
        # Initialize kernel
        kernel = self.get_kernel(device=input.device)
        
        # Load segmentation
        seg_mask = params.get('seg', None)

        # Apply convolution
        for c in self.apply_to_channel:
            channel_data = input[:, c]  # [N, ...spatial...]
            orig = channel_data.clone()
            
            if self.retain_stats:
                reduce_dims = tuple(range(1, channel_data.dim()))
                # store per-sample mean/std (shape [N])
                orig_means = channel_data.mean(dim=reduce_dims)
                orig_stds = channel_data.std(dim=reduce_dims)
            
            if self.kernel_type in ['Laplace', 'GaussianBlur']:
                x = apply_convolution(channel_data, kernel, dim=3)
            elif self.kernel_type == 'UnsharpMask':
                # blur selected channel, compute mask and add scaled mask back Isharp​=I+α(I−G​∗I)
                blurred = apply_convolution(channel_data, kernel, dim=3)
                mask = channel_data - blurred
                x = channel_data + self.unsharp_amount * mask
            elif self.kernel_type == 'Scharr':
                tot_ = torch.zeros_like(channel_data, device=input.device)
                for k in kernel:
                    if self.absolute:
                        tot_ += torch.abs(apply_convolution(channel_data, k, dim=3))
                    else:
                        tot_ += apply_convolution(channel_data, k, dim=3)
                x = tot_
            elif self.kernel_type == 'RandConv':
                # RandConv kernels are per-sample, per-call
                out = []
                for b in range(channel_data.shape[0]):
                    kernel = self.get_kernel(device=input.device)

                    conv = apply_convolution(channel_data[b:b+1], kernel, dim=3).squeeze(0)

                    if torch.rand(1).item() < self.mix_prob:
                        alpha = torch.rand(1, device=input.device)
                        conv = alpha * channel_data[b] + (1 - alpha) * conv

                    out.append(conv)

                x = torch.stack(out, dim=0)
            
            if self.retain_stats:
                # Adjust mean and std to match original
                eps = 1e-8
                reduce_dims = tuple(range(1, x.dim()))
                new_mean = x.mean(dim=reduce_dims)  # [N]
                new_std = x.std(dim=reduce_dims)    # [N]
                # reshape stats to broadcast over spatial dims: [N,1,1,...]
                shape = [x.shape[0]] + [1] * (x.dim() - 1)
                nm = new_mean.view(shape)
                ns = new_std.view(shape)
                om = orig_means.view(shape)
                os = orig_stds.view(shape)
                x = (x - nm) / (ns + eps) * os + om
            
            # Apply region selection
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                x = _apply_region_mode(orig, x, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            
            # Final safety: check if nan/inf appeared
            if torch.isnan(x).any() or torch.isinf(x).any():
                print(f"Warning nan: {self.__class__.__name__} with kernel={self.kernel_type}", flush=True)
                continue
            input[:, c] = x
        
        return input

def apply_convolution(img: torch.Tensor, kernel: torch.Tensor, dim: int) -> torch.Tensor:
    '''
    Based on https://github.com/pytorch/vision/blob/e3b5d3a8bf5e8636462fd8bce9897bccc690b2a0/torchvision/transforms/_functional_tensor.py#L746
    '''
    if not (isinstance(img, torch.Tensor)):
        raise TypeError(f"img should be Tensor. Got {type(img)}")

    if dim == 2:
        kernel = kernel.expand(img.shape[-(1 + dim)], 1, kernel.shape[0], kernel.shape[1])
        padding = [kernel.shape[2] // 2, kernel.shape[2] // 2, kernel.shape[3] // 2, kernel.shape[3] // 2]
    elif dim == 3:
        kernel = kernel.expand(img.shape[-(1 + dim)], 1, kernel.shape[0], kernel.shape[1], kernel.shape[2])
        padding = [kernel.shape[2] // 2, kernel.shape[2] // 2, kernel.shape[3] // 2, kernel.shape[3] // 2] + [kernel.shape[4] // 2, kernel.shape[4] // 2]
    else:
        raise ValueError(f"Only 2D and 3D convolution are supported. Got {dim}D.")
    
    img, need_cast, need_squeeze, out_dtype = F_t._cast_squeeze_in(img, [kernel.dtype])

    # padding = (left, right, top, bottom)
    img = F.pad(img, padding, mode="reflect")
    if dim == 2:
        img = F.conv2d(img, kernel, groups=img.shape[-(1 + dim)])
    else:  # dim == 3
        img = F.conv3d(img, kernel, groups=img.shape[-(1 + dim)])

    img = F_t._cast_squeeze_out(img, need_cast, need_squeeze, out_dtype)
    return img

def get_gaussian_kernel1d(kernel_size: int, sigma: float, dtype: torch.dtype, device: torch.device) -> Tensor:
    """Create a 1D Gaussian kernel."""

    x = torch.arange(kernel_size, dtype=dtype, device=device)
    pdf = torch.exp(-0.5 * (x / sigma).pow(2))
    kernel1d = pdf / pdf.sum()

    return kernel1d

def get_gaussian_kernel3d(kernel_size: int, sigma: torch.Tensor, dtype: torch.dtype, device: torch.device) -> Tensor:
    """
    Create a 3D Gaussian kernel by multiplying 1D kernels along each axis.
    Args:
        kernel_size (int)
        sigma (float or tuple of three floats): Standard deviation of the Gaussian kernel.
    """
    if isinstance(sigma, (int, float)):
        sigma = torch.tensor([sigma, sigma, sigma], device=device)
    elif isinstance(sigma, torch.Tensor):
        assert sigma.shape == (3,), "Sigma must be a float or a tensor of three floats."
    else:
        raise TypeError("Sigma must be a float or a tensor of three floats.")

    gz = get_gaussian_kernel1d(kernel_size, sigma[0], dtype, device)
    gy = get_gaussian_kernel1d(kernel_size, sigma[1], dtype, device)
    gx = get_gaussian_kernel1d(kernel_size, sigma[2], dtype, device)

    # Outer product using broadcasting
    kernel = gz[:, None, None] * gy[None, :, None] * gx[None, None, :]

    # Normalize
    kernel /= kernel.sum()

    return kernel

## Noise transform
class RandomGaussianNoiseGPU(ImageOnlyTransform):
    """Add random Gaussian noise to image.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        mean (float): Mean of the Gaussian noise. Default is 0.0.
        std (float): Standard deviation of the Gaussian noise. Default is 0.1.
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with added Gaussian noise.
    """
    
    def __init__(
        self,
        mean: float = 0.0,
        std: float = 0.1,
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.apply_to_channel = apply_to_channel
        self.mean = mean
        self.std = std
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:
        # Generate Gaussian noise with the same shape as input
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            if self.same_on_batch:
                std = torch.rand(1, device=input.device, dtype=input.dtype) * self.std
                noise = torch.randn_like(input[:,c], device=input.device, dtype=input.dtype)
                noise = noise * std + self.mean
            else:
                std = torch.rand(input.shape[0], device=input.device, dtype=input.dtype) * self.std
                noise = torch.randn_like(input[:,c], device=input.device, dtype=input.dtype)
                for i in range(input.shape[0]):
                    noise[i] = noise[i] * std[i] + self.mean
            
            orig = input[:, c]
            x = orig + noise
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                x = _apply_region_mode(orig, x, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(x).any() or torch.isinf(x).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = x
        
        return input

## Multiplicative brightness transform
class RandomBrightnessGPU(ImageOnlyTransform):
    """Apply random brightness adjustment to image.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        brightness_range (tuple of float): Range of brightness multipliers. Default is (0.9, 1.1).
        apply_to_channel (list of int): List of channel indices to apply the brightness adjustment to. Default is [0].
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with adjusted brightness.
    """
    
    def __init__(
        self,
        brightness_range: list[float, float] = (0.9, 1.1),
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.brightness_range = brightness_range
        self.apply_to_channel = apply_to_channel
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:

        # Apply brightness adjustment
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            channel_data = input[:, c]  # [N, ...spatial...]
            orig = channel_data.clone()
            if self.same_on_batch:
                factor = torch.rand(1, device=input.device, dtype=input.dtype) * (self.brightness_range[1] - self.brightness_range[0]) + self.brightness_range[0]
                x = channel_data * factor
            else:
                factor = torch.rand(input.shape[0], device=input.device, dtype=input.dtype) * (self.brightness_range[1] - self.brightness_range[0]) + self.brightness_range[0]
                x = channel_data.clone()
                for i in range(input.shape[0]):
                    x[i] = x[i] * factor[i]
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                x = _apply_region_mode(orig, x, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(x).any() or torch.isinf(x).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = x
        
        return input

## Gamma transform
class RandomGammaGPU(ImageOnlyTransform):
    """Apply random gamma adjustment to image.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        gamma_range (tuple of float): Range of gamma multipliers. Default is (0.9, 1.1).
        invert_image (bool): If True, invert the image before and after gamma adjustment. Default is False.
        apply_to_channel (list of int): List of channel indices to apply the gamma adjustment to. Default is [0].
        retain_stats (bool): If True, retain the original mean and standard deviation of the image after gamma adjustment. Default is False.
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with adjusted brightness.
    """
    
    def __init__(
        self,
        gamma_range: list[float, float] = (0.9, 1.1),
        invert_image: bool = False,
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        retain_stats: bool = False,
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.gamma_range = gamma_range
        self.invert_image = invert_image
        self.retain_stats = retain_stats
        self.apply_to_channel = apply_to_channel
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:

        # Apply gamma transform
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            if self.invert_image:
                channel_data = -input[:, c]  # [N, ...spatial...]
            else:
                channel_data = input[:, c]  # [N, ...spatial...]
            orig_full = input[:, c]
            
            if self.retain_stats:
                reduce_dims = tuple(range(1, channel_data.dim()))
                # store per-sample mean/std (shape [N])
                orig_means = channel_data.mean(dim=reduce_dims)
                orig_stds = channel_data.std(dim=reduce_dims)
            
            if self.same_on_batch:
                gamma = torch.rand(1, device=input.device, dtype=input.dtype) * (self.gamma_range[1] - self.gamma_range[0]) + self.gamma_range[0]
            else:
                gamma = torch.rand(input.shape[0], device=input.device, dtype=input.dtype) * (self.gamma_range[1] - self.gamma_range[0]) + self.gamma_range[0]
            
            # Compute min and range per batch element for the current channel
            # Flatten spatial dimensions to compute min/max per batch element
            batch_size = channel_data.shape[0]
            flat_data = channel_data.view(batch_size, -1)  # [N, spatial_flattened]
            minm = flat_data.min(dim=1, keepdim=self.keepdim)[0]  # [N, 1]
            maxm = flat_data.max(dim=1, keepdim=self.keepdim)[0]  # [N, 1]
            rnge = maxm - minm
            
            # Reshape min, max, range to broadcast over spatial dims: [N, 1] -> [N, 1, 1, ...]
            reshape_dims = [batch_size] + [1] * (channel_data.dim() - 1)
            minm = minm.view(reshape_dims)
            rnge = rnge.view(reshape_dims)
            
            # Reshape gamma to broadcast properly: [N] -> [N, 1, 1, ...]
            if not self.same_on_batch:
                gamma = gamma.view(reshape_dims)
            
            # Apply gamma transform per batch element
            channel_data = torch.pow(((channel_data - minm) / (rnge + 1e-8)), gamma) * rnge + minm
            
            if self.retain_stats:
                # Adjust mean and std to match original
                eps = 1e-8
                reduce_dims = tuple(range(1, channel_data.dim()))
                new_mean = channel_data.mean(dim=reduce_dims)  # [N]
                new_std = channel_data.std(dim=reduce_dims)    # [N]
                # reshape stats to broadcast over spatial dims: [N,1,1,...]
                shape = [channel_data.shape[0]] + [1] * (channel_data.dim() - 1)
                nm = new_mean.view(shape)
                ns = new_std.view(shape)
                om = orig_means.view(shape)
                os = orig_stds.view(shape)
                channel_data = (channel_data - nm) / (ns + eps) * os + om
            
            if self.invert_image:
                channel_data = -channel_data
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                channel_data = _apply_region_mode(orig_full, channel_data, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(channel_data).any() or torch.isinf(channel_data).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = channel_data
        
        return input

## nnunetv2 contrast transform
class RandomContrastGPU(ImageOnlyTransform):
    """Apply random gamma adjustment to image.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        contrast_range (tuple of float): Range of gamma multipliers. Default is (0.9, 1.1).
        apply_to_channel (list of int): List of channel indices to apply the gamma adjustment to. Default is [0].
        retain_stats (bool): If True, retain the original mean and standard deviation of the image after gamma adjustment. Default is False.
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with adjusted brightness.
    """
    
    def __init__(
        self,
        contrast_range: list[float, float] = (0.9, 1.1),
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        retain_stats: bool = False,
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.contrast_range = contrast_range
        self.apply_to_channel = apply_to_channel
        self.retain_stats = retain_stats
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:

        # Apply brightness adjustment
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            channel_data = input[:, c]  # [N, ...spatial...]
            orig = channel_data.clone()
            if self.retain_stats:
                reduce_dims = tuple(range(1, channel_data.dim()))
                # store per-sample mean/std (shape [N])
                orig_means = channel_data.mean(dim=reduce_dims)
                orig_stds = channel_data.std(dim=reduce_dims)
            
            if self.same_on_batch:
                factor = torch.rand(1, device=input.device, dtype=input.dtype) * (self.contrast_range[1] - self.contrast_range[0]) + self.contrast_range[0]
                x = channel_data.clone()
                for i in range(input.shape[0]):
                    mean = x[i].mean()
                    x[i] = (x[i] - mean) * factor + mean
            else:
                factor = torch.rand(input.shape[0], device=input.device, dtype=input.dtype) * (self.contrast_range[1] - self.contrast_range[0]) + self.contrast_range[0]
                x = channel_data.clone()
                for i in range(input.shape[0]):
                    mean = x[i].mean()
                    x[i] = (x[i] - mean) * factor[i] + mean
            
            if self.retain_stats:
                # Adjust mean and std to match original
                eps = 1e-8
                reduce_dims = tuple(range(1, x.dim()))
                new_mean = x.mean(dim=reduce_dims)  # [N]
                new_std = x.std(dim=reduce_dims)    # [N]
                # reshape stats to broadcast over spatial dims: [N,1,1,...]
                shape = [x.shape[0]] + [1] * (x.dim() - 1)
                nm = new_mean.view(shape)
                ns = new_std.view(shape)
                om = orig_means.view(shape)
                os = orig_stds.view(shape)
                x = (x - nm) / (ns + eps) * os + om
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                x = _apply_region_mode(orig, x, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(x).any() or torch.isinf(x).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = x
        
        return input

## Function transform
class RandomFunctionGPU(ImageOnlyTransform):
    """Apply function to the image based on probability.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        func (callable): Random function to apply. Default is a gamma adjustment function.
        apply_to_channel (list of int): List of channel indices to apply the function to. Default is [0].
        retain_stats (bool): If True, retain the original mean and standard deviation of the image after gamma adjustment. Default is False.
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with adjusted brightness.
    """
    
    def __init__(
        self,
        func: callable = lambda x: x ** 2,
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        retain_stats: bool = False,
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.func = func
        self.retain_stats = retain_stats
        self.apply_to_channel = apply_to_channel
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:

        # Apply function transform
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            x = input[:, c]  # shape [N, ...spatial...]
            orig = x.clone()
            if self.retain_stats:
                reduce_dims = tuple(range(1, x.dim()))
                # store per-sample mean/std (shape [N])
                orig_means = x.mean(dim=reduce_dims)
                orig_stds = x.std(dim=reduce_dims)
            
            # Normalize to make values >=0
            x = (x - x.min()) / (x.max() - x.min() + 0.00001)

            # Apply function
            x = self.func(x)

            if self.retain_stats:
                # Adjust mean and std to match original
                eps = 1e-8
                reduce_dims = tuple(range(1, x.dim()))
                new_mean = x.mean(dim=reduce_dims)  # [N]
                new_std = x.std(dim=reduce_dims)    # [N]
                # reshape stats to broadcast over spatial dims: [N,1,1,...]
                shape = [x.shape[0]] + [1] * (x.dim() - 1)
                nm = new_mean.view(shape)
                ns = new_std.view(shape)
                om = orig_means.view(shape)
                os = orig_stds.view(shape)
                x = (x - nm) / (ns + eps) * os + om
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                x = _apply_region_mode(orig, x, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(x).any() or torch.isinf(x).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = x
        
        return input

## Inverse transform
class RandomInverseGPU(ImageOnlyTransform):
    """Inverse image based on probability.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        apply_to_channel (list of int): List of channel indices to apply the brightness adjustment to. Default is [0].
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with adjusted brightness.
    """
    
    def __init__(
        self,
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        retain_stats: bool = False,
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.apply_to_channel = apply_to_channel
        self.retain_stats = retain_stats
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:

        # Inverse image
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            for i in range(input.shape[0]):
                x= input[i, c]  # shape [...spatial...]
                orig = x.clone()
                if self.retain_stats:
                    orig_means = x.mean()
                    orig_stds = x.std()
                max_val = x.max()
                x = max_val - x
                if self.retain_stats:
                    # Adjust mean and std to match original
                    eps = 1e-8
                    new_mean = x.mean()  # scalar
                    new_std = x.std()    # scalar
                    x = (x - new_mean) / (new_std + eps) * orig_stds + orig_means
                if not seg_mask is None:
                    region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask[i])
                    x = _apply_region_mode(orig, x, seg_mask[i], region_mode, mix_in_out=self.mix_in_out)
                # Final safety: check if nan/inf appeared
                if torch.isnan(x).any() or torch.isinf(x).any():
                    print(f"Warning nan: {self.__class__.__name__}", flush=True)
                    continue
                input[i, c] = x
        
        return input

## Histogram transform
class RandomHistogramEqualizationGPU(ImageOnlyTransform):
    """Apply histogram equalization transformation to the image based on probability.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        apply_to_channel (list of int): List of channel indices to apply the histogram equalization to. Default is [0].
        retain_stats (bool): If True, retain the original mean and standard deviation of the image after histogram equalization. Default is False.
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with adjusted brightness.
    """
    
    def __init__(
        self,
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        retain_stats: bool = False,
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.retain_stats = retain_stats
        self.apply_to_channel = apply_to_channel
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:

        # Apply histogram equalization transform
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            channel_data = input[:, c]  # shape [N, ...spatial...]
            orig = channel_data.clone()
            
            if self.retain_stats:
                reduce_dims = tuple(range(1, channel_data.dim()))
                # store per-sample mean/std (shape [N])
                orig_means = channel_data.mean(dim=reduce_dims)
                orig_stds = channel_data.std(dim=reduce_dims)
            
            # Process each batch element independently
            batch_size = channel_data.shape[0]
            for b in range(batch_size):
                img_b = channel_data[b]  # Single image from batch [...spatial...]
                
                img_min, img_max = img_b.min(), img_b.max()
                
                # Flatten the image and compute the histogram
                img_flattened = img_b.flatten().to(torch.float32)
                hist = torch.histc(img_flattened, bins=256, min=img_min.item(), max=img_max.item())
                
                # Compute the normalized cumulative distribution function (CDF)
                cdf = hist.cumsum(dim=0)
                cdf_min = cdf[cdf > 0].min() if (cdf > 0).any() else cdf.min()
                cdf = (cdf - cdf_min) / (cdf[-1] - cdf_min + 1e-10)  # Normalize to [0,1]
                cdf = cdf * (img_max - img_min) + img_min  # Scale back to image range
                
                # Compute bin edges and indices
                bin_width = (img_max - img_min) / 256
                indices = ((img_flattened - img_min) / (bin_width + 1e-10)).long()
                indices = torch.clamp(indices, 0, 255)
                
                # Perform histogram equalization
                img_eq = cdf[indices]
                channel_data[b] = img_eq.reshape(img_b.shape)
            
            if self.retain_stats:
                # Adjust mean and std to match original
                eps = 1e-8
                reduce_dims = tuple(range(1, channel_data.dim()))
                new_mean = channel_data.mean(dim=reduce_dims)  # [N]
                new_std = channel_data.std(dim=reduce_dims)    # [N]
                # reshape stats to broadcast over spatial dims: [N,1,1,...]
                shape = [channel_data.shape[0]] + [1] * (channel_data.dim() - 1)
                nm = new_mean.view(shape)
                ns = new_std.view(shape)
                om = orig_means.view(shape)
                os = orig_stds.view(shape)
                channel_data = (channel_data - nm) / (ns + eps) * os + om
            
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                channel_data = _apply_region_mode(orig, channel_data, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(channel_data).any() or torch.isinf(channel_data).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = channel_data
        
        return input


class RandomBiasFieldGPU(ImageOnlyTransform):
    """Apply a smooth multiplicative bias field to selected channels.

    The bias field simulates low-frequency intensity inhomogeneity (MRI bias field).
    It is constructed as the exponential of a polynomial combination of the
    spatial coordinates (x, y, z) up to a given order with random coefficients.

    Supports 2D (N, C, H, W) and 3D (N, C, D, H, W) tensors.

    Args:
        coefficients (float | tuple[float, float]): If float c, coefficients sampled
            uniformly from (-c, c). If tuple (a, b) coefficients sampled from (a, b).
        order (int): Polynomial order (>=0).
        apply_to_channel (list[int]): Channels to which the bias field is applied.
        invert (bool): If True, uses inverse bias field (1 / field).
        retain_stats (bool): If True, restores original per-sample mean and std for affected channels.
        same_on_batch (bool): If True, uses the same sampled coefficients for all batch elements.
        p (float): Application probability.
        keepdim (bool): Keep input dimensions flag (passed to base).
    """

    def __init__(
        self,
        coefficients: Union[float, Tuple[float, float]] = 0.5,
        order: int = 3,
        apply_to_channel: list[int] = [0],
        invert: bool = False,
        retain_stats: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        same_on_batch: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        if isinstance(coefficients, (int, float)):
            self.coeff_range = (-float(coefficients), float(coefficients))
        elif isinstance(coefficients, (tuple, list)) and len(coefficients) == 2:
            self.coeff_range = (float(coefficients[0]), float(coefficients[1]))
        else:
            raise TypeError("coefficients must be float or (min, max) tuple")
        if not isinstance(order, int) or order < 0:
            raise ValueError("order must be a non-negative int")
        self.order = order
        self.apply_to_channel = apply_to_channel
        self.invert = invert
        self.retain_stats = retain_stats
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    def _num_coeffs(self, dim: int) -> int:
        # Count coefficients generated by nested loops matching TorchIO logic.
        count = 0
        if dim == 3:
            for xo in range(self.order + 1):
                for yo in range(self.order + 1 - xo):
                    for zo in range(self.order + 1 - (xo + yo)):
                        count += 1
        elif dim == 2:
            for xo in range(self.order + 1):
                for yo in range(self.order + 1 - xo):
                    count += 1
        else:
            raise ValueError("Only 2D or 3D spatial dims supported for bias field")
        return count

    def _sample_coeffs(self, batch_size: int, device: torch.device, dtype: torch.dtype, dim: int) -> torch.Tensor:
        n = self._num_coeffs(dim)
        low, high = self.coeff_range
        if self.same_on_batch:
            coeff = torch.empty(n, 1, device=device, dtype=dtype).uniform_(low, high)
            coeff = coeff.expand(n, batch_size)
        else:
            coeff = torch.empty(n, batch_size, device=device, dtype=dtype).uniform_(low, high)
        return coeff  # shape (n_coeffs, B)

    def _make_grids(self, spatial_shape: Tuple[int, ...], device: torch.device, dtype: torch.dtype) -> List[torch.Tensor]:
        # Create coordinate grids normalized to [-1, 1]
        if len(spatial_shape) == 2:
            h, w = spatial_shape
            ys = torch.linspace(-1, 1, h, device=device, dtype=dtype)
            xs = torch.linspace(-1, 1, w, device=device, dtype=dtype)
            y_grid, x_grid = torch.meshgrid(ys, xs, indexing='ij')
            return [x_grid, y_grid]
        elif len(spatial_shape) == 3:
            d, h, w = spatial_shape
            zs = torch.linspace(-1, 1, d, device=device, dtype=dtype)
            ys = torch.linspace(-1, 1, h, device=device, dtype=dtype)
            xs = torch.linspace(-1, 1, w, device=device, dtype=dtype)
            z_grid, y_grid, x_grid = torch.meshgrid(zs, ys, xs, indexing='ij')
            return [x_grid, y_grid, z_grid]
        else:
            raise ValueError("Spatial dims must be 2 or 3 for bias field")

    @torch.no_grad()
    def apply_transform(
        self,
        input: Tensor,
        params: Dict[str, Tensor],
        flags: Dict[str, Any],
        transform: Optional[Tensor] = None,
    ) -> Tensor:
        # input: (N, C, [D,] H, W)
        if input.dim() not in (4, 5):
            raise ValueError("Expected 4D or 5D tensor (N,C,...) for RandomBiasFieldGPU")
        batch_size = input.shape[0]
        spatial = input.shape[2:]
        dim = len(spatial)
        device = input.device
        dtype = input.dtype

        coeffs = self._sample_coeffs(batch_size, device, dtype, dim)  # (n_coeffs, B)
        grids = self._make_grids(spatial, device, dtype)
        seg_mask = params.get('seg', None)

        # Initialize bias map per batch element
        bias_map = torch.zeros((batch_size, *spatial), device=device, dtype=dtype)

        idx = 0
        if dim == 3:
            xg, yg, zg = grids  # each shape (D,H,W)
            for xo in range(self.order + 1):
                x_term = xg.pow(xo) if xo > 0 else 1.0
                for yo in range(self.order + 1 - xo):
                    y_term = yg.pow(yo) if yo > 0 else 1.0
                    for zo in range(self.order + 1 - (xo + yo)):
                        z_term = zg.pow(zo) if zo > 0 else 1.0
                        # term shape (D,H,W)
                        term = x_term * y_term * z_term
                        # Add coefficient * term for each batch element
                        bias_map += coeffs[idx].view(-1, *([1] * dim)) * term  # broadcast over spatial
                        idx += 1
        else:  # dim == 2
            xg, yg = grids  # (H,W)
            for xo in range(self.order + 1):
                x_term = xg.pow(xo) if xo > 0 else 1.0
                for yo in range(self.order + 1 - xo):
                    y_term = yg.pow(yo) if yo > 0 else 1.0
                    term = x_term * y_term  # (H,W)
                    bias_map += coeffs[idx].view(-1, *([1] * dim)) * term
                    idx += 1

        # Exponential to ensure positive field
        bias_field = torch.exp(bias_map)  # (N, *spatial)
        if self.invert:
            bias_field = 1.0 / (bias_field + 1e-8)

        # Apply to channels
        for c in self.apply_to_channel:
            if c < 0 or c >= input.shape[1]:
                continue  # skip invalid channel index
            channel = input[:, c]
            orig = channel.clone()
            if self.retain_stats:
                reduce_dims = tuple(range(1, channel.dim()))
                orig_mean = channel.mean(dim=reduce_dims)
                orig_std = channel.std(dim=reduce_dims)
            channel = channel * bias_field
            if self.retain_stats:
                eps = 1e-8
                new_mean = channel.mean(dim=reduce_dims)
                new_std = channel.std(dim=reduce_dims)
                # reshape stats for broadcasting
                shape = [channel.shape[0]] + [1] * (channel.dim() - 1)
                om = orig_mean.view(shape)
                os = orig_std.view(shape)
                nm = new_mean.view(shape)
                ns = new_std.view(shape)
                channel = (channel - nm) / (ns + eps) * os + om
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                channel = _apply_region_mode(orig, channel, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(channel).any() or torch.isinf(channel).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = channel
        
        return input

# Random clamping transform
class RandomClampGPU(ImageOnlyTransform):
    """Apply random gamma adjustment to image.
    If the image is torch Tensor, it is expected to have [N, C, X, Y] or [N, C, X, Y, Z] shape.

    Args:
        max_clamp_amount (float): Amount to clamp the image values (0 < min_clamp < max_clamp_amount and 1 - max_clamp_amount < max_clamp < 1). Default is 0.2.
        apply_to_channel (list of int): List of channel indices to apply the gamma adjustment to. Default is [0].
        retain_stats (bool): If True, retain the original mean and standard deviation of the image after gamma adjustment. Default is False.
        same_on_batch (bool): Apply the same transformation across the batch. Default is False.
        p (float): Probability of applying the transform. Default is 1.0.
        keepdim (bool): Whether to keep the number of dimensions. Default is False.

    Returns:
        Tensor: Image with adjusted brightness.
    """
    
    def __init__(
        self,
        max_clamp_amount: float = 0.2,
        apply_to_channel: list[int] = [0],  # Apply to first channel by default
        retain_stats: bool = False,
        same_on_batch: bool = False,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        mix_in_out: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.max_clamp_amount = max_clamp_amount
        self.apply_to_channel = apply_to_channel
        self.retain_stats = retain_stats
        self.in_seg = in_seg
        self.out_seg = out_seg
        self.mix_in_out = mix_in_out

    @torch.no_grad()  # disable gradients for efficiency
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:

        # Apply clamping
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            channel_data = input[:, c]  # [N, ...spatial...]
            orig = channel_data.clone()
            if self.retain_stats:
                reduce_dims = tuple(range(1, channel_data.dim()))
                # store per-sample mean/std (shape [N])
                orig_means = channel_data.mean(dim=reduce_dims)
                orig_stds = channel_data.std(dim=reduce_dims)
            
            if self.same_on_batch:
                min_percentile = torch.rand(1, device=input.device, dtype=input.dtype) * self.max_clamp_amount
                max_percentile = 1.0 - (torch.rand(1, device=input.device, dtype=input.dtype) * self.max_clamp_amount)
                x = channel_data.clone()
                for i in range(input.shape[0]):
                    min_val = torch.quantile(x[i].flatten(), min_percentile)
                    max_val = torch.quantile(x[i].flatten(), max_percentile)
                    x[i] = torch.clamp(x[i], min_val, max_val)
            else:
                x = channel_data.clone()
                for i in range(input.shape[0]):
                    min_percentile = torch.rand(1, device=input.device, dtype=input.dtype) * self.max_clamp_amount
                    max_percentile = 1.0 - (torch.rand(1, device=input.device, dtype=input.dtype) * self.max_clamp_amount)
                    min_val = torch.quantile(x[i].flatten(), min_percentile)
                    max_val = torch.quantile(x[i].flatten(), max_percentile)
                    x[i] = torch.clamp(x[i], min_val, max_val)
            
            if self.retain_stats:
                # Adjust mean and std to match original
                eps = 1e-8
                reduce_dims = tuple(range(1, x.dim()))
                new_mean = x.mean(dim=reduce_dims)  # [N]
                new_std = x.std(dim=reduce_dims)    # [N]
                # reshape stats to broadcast over spatial dims: [N,1,1,...]
                shape = [x.shape[0]] + [1] * (x.dim() - 1)
                nm = new_mean.view(shape)
                ns = new_std.view(shape)
                om = orig_means.view(shape)
                os = orig_stds.view(shape)
                x = (x - nm) / (ns + eps) * os + om
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                x = _apply_region_mode(orig, x, seg_mask, region_mode, mix_in_out=self.mix_in_out)
            # Final safety: check if nan/inf appeared
            if torch.isnan(x).any() or torch.isinf(x).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = x
        
        return input


class ZscoreNormalizationGPU(ImageOnlyTransform):
    """Apply z-score normalization to selected channels.

    Args:
        apply_to_channel (list[int]): Channels to which the normalization is applied.
        p (float): Application probability.
        keepdim (bool): Keep input dimensions flag (passed to base).
    """

    def __init__(
        self,
        apply_to_channel: list[int] = [0],
        keepdim: bool = True,
        in_seg: float = 0.0,
        out_seg: float = 0.0,
        p: float = 1.0,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=False, keepdim=keepdim)
        self.apply_to_channel = apply_to_channel
        self.in_seg = in_seg
        self.out_seg = out_seg

    @torch.no_grad()
    def apply_transform(
        self,
        input: Tensor,
        params: Dict[str, Tensor],
        flags: Dict[str, Any],
        transform: Optional[Tensor] = None,
    ) -> Tensor:
        # input: (N, C, [D,] H, W)
        seg_mask = params.get('seg', None)
        for c in self.apply_to_channel:
            if c < 0 or c >= input.shape[1]:
                continue  # skip invalid channel index
            channel = input[:, c]
            orig = channel.clone()
            reduce_dims = tuple(range(1, channel.dim()))
            mean = channel.mean(dim=reduce_dims, keepdim=True)
            # use unbiased=False for stability, and clamp std to avoid division by ~0
            std = channel.std(dim=reduce_dims, keepdim=True, unbiased=False).clamp_min(1e-8)
            channel = (channel - mean) / std
            if not seg_mask is None:
                region_mode = _choose_region_mode(self.in_seg, self.out_seg, seg_mask)
                channel = _apply_region_mode(orig, channel, seg_mask, region_mode)
            # Final safety: check if nan/inf appeared
            if torch.isnan(channel).any() or torch.isinf(channel).any():
                print(f"Warning nan: {self.__class__.__name__}", flush=True)
                continue
            input[:, c] = channel
        
        return input




import os
import json
import torch
import numpy as np
from typing import Union, Tuple

from batchgeneratorsv2.helpers.scalar_type import RandomScalar
from batchgeneratorsv2.transforms.intensity.brightness import MultiplicativeBrightnessTransform
from batchgeneratorsv2.transforms.intensity.contrast import ContrastTransform, BGContrast
from batchgeneratorsv2.transforms.intensity.gamma import GammaTransform
from batchgeneratorsv2.transforms.intensity.gaussian_noise import GaussianNoiseTransform
from batchgeneratorsv2.transforms.noise.gaussian_blur import GaussianBlurTransform
from batchgeneratorsv2.transforms.spatial.low_resolution import SimulateLowResolutionTransform
from batchgeneratorsv2.transforms.spatial.mirroring import MirrorTransform
from batchgeneratorsv2.transforms.spatial.spatial import SpatialTransform
from batchgeneratorsv2.transforms.utils.random import RandomTransform
from batchgeneratorsv2.transforms.utils.compose import ComposeTransforms
from batchgeneratorsv2.transforms.utils.pseudo2d import Convert3DTo2DTransform, Convert2DTo3DTransform

from auglab.transforms.cpu.artifact import ArtifactTransform
from auglab.transforms.cpu.contrast import ConvTransform, HistogramEqualTransform, FunctionTransform
from auglab.transforms.cpu.fromSeg import RedistributeTransform
from auglab.transforms.cpu.spatial import SpatialCustomTransform, ShapeTransform

class AugTransforms(ComposeTransforms):
    def __init__(self, json_path: str, do_dummy_2d_data_aug: bool, patch_size: Union[np.ndarray, Tuple[int]],
                 rotation_for_DA: RandomScalar, mirror_axes: Tuple[int]):
        # Load transform parameters from JSON
        config_path = os.path.join(json_path)
        with open(config_path, 'r') as f:
            config = json.load(f)
        
        if 'CPU' in config.keys():
            self.transform_params = config['CPU']
        else:
            self.transform_params = config
            
        self.transforms = self._build_transforms(
            do_dummy_2d_data_aug=do_dummy_2d_data_aug,
            patch_size=patch_size,
            rotation_for_DA=rotation_for_DA,
            mirror_axes=mirror_axes
        )
        super().__init__(transforms=self.transforms)

    def _build_transforms(self, do_dummy_2d_data_aug: bool, patch_size: Union[np.ndarray, Tuple[int]],
                          rotation_for_DA: RandomScalar, mirror_axes: Tuple[int]):
        transform_params = self.transform_params
        transforms = []

        # Scharr filter
        conv_params = transform_params.get('ConvTransform')
        if conv_params is not None:
            transforms.append(RandomTransform(
                ConvTransform(
                    kernel_type=conv_params.get('kernel_type', 'Scharr'),
                    absolute=conv_params.get('absolute', True),
                    retain_stats=transform_params.get('retain_stats', False)
                ), apply_probability=conv_params.get('probability', 0)
            ))

        # Apply functions
        func_list = [
            lambda x: torch.log(1 + x),
            torch.sqrt,
            torch.sin,
            torch.exp,
            lambda x: 1/(1 + torch.exp(-x)),
        ]
        func_params = transform_params.get('FunctionTransform')
        if func_params is not None:
            for func in func_list:
                transforms.append(RandomTransform(
                    FunctionTransform(
                        function=func,
                        retain_stats=transform_params.get('retain_stats', False)
                    ), apply_probability=func_params.get('probability', 0)
                ))
        
        # Histogram manipulations
        hist_params = transform_params.get('HistogramEqualTransform')
        if hist_params is not None:
            transforms.append(RandomTransform(
                HistogramEqualTransform(
                    retain_stats=transform_params.get('retain_stats', False)
                ), apply_probability=hist_params.get('probability', 0)
            ))

        # Redistribute segmentation values
        redist_params = transform_params.get('RedistributeTransform')
        if redist_params is not None:
            transforms.append(RandomTransform(
                RedistributeTransform(
                    in_seg=redist_params.get('in_seg', 0),
                    retain_stats=transform_params.get('retain_stats', False)
                ), apply_probability=redist_params.get('probability', 0)
            ))

        # Resolution transforms
        shape_params = transform_params.get('ShapeTransform')
        if shape_params is not None:
            transforms.append(RandomTransform(
                ShapeTransform(
                    shape_min=shape_params.get('shape_min'),
                    ignore_axes=tuple(shape_params.get('ignore_axes', None)) if shape_params.get('ignore_axes', None) is not None else None,
                ), apply_probability=shape_params.get('probability', 0)
        ))

        # Artifacts generation
        artifact_params = transform_params.get('ArtifactTransform')
        if artifact_params is not None:
            transforms.append(RandomTransform(
                ArtifactTransform(
                    motion=artifact_params.get('motion', False),
                    ghosting=artifact_params.get('ghosting', False),
                    spike=artifact_params.get('spike', False),
                    bias_field=artifact_params.get('bias_field', False),
                    blur=artifact_params.get('blur', False),
                    noise=artifact_params.get('noise', False),
                    swap=artifact_params.get('swap', False),
                    random_pick=artifact_params.get('random_pick', False)
                ), apply_probability=artifact_params.get('probability', 0)
            ))

        # Spatial transforms
        spatial_custom_params = transform_params.get('SpatialCustomTransform')
        if spatial_custom_params is not None:
            transforms.append(RandomTransform(
                SpatialCustomTransform(
                    flip=spatial_custom_params.get('flip', False),
                    affine=spatial_custom_params.get('affine', False),
                    elastic=spatial_custom_params.get('elastic', False),
                    anisotropy=spatial_custom_params.get('anisotropy', False),
                    random_pick=spatial_custom_params.get('random_pick', False)
                ), apply_probability=spatial_custom_params.get('probability', 0)
            ))
        
        # Spatial nnunet transform
        if do_dummy_2d_data_aug:
            ignore_axes = (0,)
            transforms.append(Convert3DTo2DTransform())
            patch_size_spatial = patch_size[1:]
        else:
            patch_size_spatial = patch_size
            ignore_axes = None
        
        spatial_params = transform_params.get('SpatialTransform')
        if spatial_params is not None:
            transforms.append(
                SpatialTransform(
                    patch_size_spatial, 
                    patch_center_dist_from_border=spatial_params.get('patch_center_dist_from_border', 0), 
                    random_crop=spatial_params.get('random_crop', False),
                    p_elastic_deform=spatial_params.get('p_elastic_deform', 0),
                    p_rotation=spatial_params.get('p_rotation', 0),
                    rotation=rotation_for_DA, 
                    p_scaling=spatial_params.get('p_scaling', 0), 
                    scaling=spatial_params.get('scaling', (0.7, 1.4)), 
                    p_synchronize_scaling_across_axes=spatial_params.get('p_synchronize_scaling_across_axes', 1),
                    bg_style_seg_sampling=spatial_params.get('bg_style_seg_sampling', False),
                    mode_seg='nearest'
                )
            )

        if do_dummy_2d_data_aug:
            transforms.append(Convert2DTo3DTransform())
        
        # Noise transforms
        noise_params = transform_params.get('GaussianNoiseTransform')
        if noise_params is not None:
            transforms.append(RandomTransform(
                GaussianNoiseTransform(
                    noise_variance=tuple(noise_params.get('noise_variance', (0, 0.1))),
                    p_per_channel=noise_params.get('p_per_channel', 1),
                    synchronize_channels=noise_params.get('synchronize_channels', True)
                ), apply_probability=noise_params.get('probability', 0)
            ))

        # Gaussian blur
        blur_params = transform_params.get('GaussianBlurTransform')
        if blur_params is not None:
            transforms.append(RandomTransform(
                GaussianBlurTransform(
                    blur_sigma=tuple(blur_params.get('blur_sigma', (0.5, 1.))),
                    synchronize_channels=blur_params.get('synchronize_channels', False),
                    synchronize_axes=blur_params.get('synchronize_axes', False),
                    p_per_channel=blur_params.get('p_per_channel', 0.5),
                    benchmark=blur_params.get('benchmark', True)
                ), apply_probability=blur_params.get('probability', 0)
            ))

        # Brightness transforms
        bright_params = transform_params.get('MultiplicativeBrightnessTransform')
        if bright_params is not None:
            transforms.append(RandomTransform(
                MultiplicativeBrightnessTransform(
                    multiplier_range=BGContrast(tuple(bright_params.get('multiplier_range', (0.75, 1.25)))),
                    synchronize_channels=bright_params.get('synchronize_channels', False),
                    p_per_channel=bright_params.get('p_per_channel', 1)
                ), apply_probability=bright_params.get('probability', 0)
            ))

        # Contrast transforms
        contrast_params = transform_params.get('ContrastTransform')
        if contrast_params is not None:
            transforms.append(RandomTransform(
                ContrastTransform(
                    contrast_range=BGContrast(tuple(contrast_params.get('contrast_range', (0.75, 1.25)))),
                    preserve_range=contrast_params.get('preserve_range', True),
                    synchronize_channels=contrast_params.get('synchronize_channels', False),
                    p_per_channel=contrast_params.get('p_per_channel', 1)
                ), apply_probability=contrast_params.get('probability', 0)
            ))

        # Simulate low resolution
        lowres_params = transform_params.get('SimulateLowResolutionTransform')
        if lowres_params is not None:
            transforms.append(RandomTransform(
                SimulateLowResolutionTransform(
                    scale=tuple(lowres_params.get('scale', (0.3, 1))),
                    synchronize_channels=lowres_params.get('synchronize_channels', True),
                    synchronize_axes=lowres_params.get('synchronize_axes', False),
                    ignore_axes=tuple(lowres_params.get('ignore_axes', ())),
                    allowed_channels=lowres_params.get('allowed_channels', None),
                    p_per_channel=lowres_params.get('p_per_channel', 0.5)
                ), apply_probability=lowres_params.get('probability', 0)
            ))

        # Gamma transforms
        gamma_inv_params = transform_params.get('GammaTransform_invert')
        if gamma_inv_params is not None:
            transforms.append(RandomTransform(
                GammaTransform(
                    gamma=BGContrast(tuple(gamma_inv_params.get('gamma', (0.7, 1.5)))),
                    p_invert_image=gamma_inv_params.get('p_invert_image', 1),
                    synchronize_channels=gamma_inv_params.get('synchronize_channels', False),
                    p_per_channel=gamma_inv_params.get('p_per_channel', 1),
                    p_retain_stats=gamma_inv_params.get('p_retain_stats', 1)
                ), apply_probability=gamma_inv_params.get('probability', 0)
        ))

        gamma_params = transform_params.get('GammaTransform')
        if gamma_params is not None:
            transforms.append(RandomTransform(
                GammaTransform(
                    gamma=BGContrast(tuple(gamma_params.get('gamma', (0.7, 1.5)))),
                    p_invert_image=gamma_params.get('p_invert_image', 0),
                    synchronize_channels=gamma_params.get('synchronize_channels', False),
                    p_per_channel=gamma_params.get('p_per_channel', 1),
                    p_retain_stats=gamma_params.get('p_retain_stats', 1)
                ), apply_probability=gamma_params.get('probability', 0)
        ))

        # Mirroring transforms
        if transform_params.get('mirror_axes') is not None and len(transform_params['mirror_axes']) > 0:
            transforms.append(
                MirrorTransform(
                    allowed_axes=mirror_axes
                )
            )

        return transforms

class AugTransformsTest(ComposeTransforms):
    def __init__(self):
        self.transforms = self._build_transforms()
        super().__init__(transforms=self.transforms)

    def _build_transforms(self):
        transforms = []

        # Scharr filter
        transforms.append(RandomTransform(
            ConvTransform(
                kernel_type="Scharr",
                absolute=True,
            ), apply_probability=0.9
        ))

        # Affine transforms
        transforms.append(RandomTransform(
            SpatialCustomTransform(
                affine=True,
            ), apply_probability=0.9
        ))

        return transforms


import torch
from torch.nn import functional as F

from typing import Any, Dict, Optional, Tuple, Union, List
from kornia.core import Tensor

from auglab.transforms.gpu.base import ImageOnlyTransform


def _normal_pdf(x: torch.Tensor, mean: torch.Tensor, std: torch.Tensor) -> torch.Tensor:
    inv = 1.0 / (std + 1e-6)
    return (inv / (torch.sqrt(torch.tensor(2.0 * 3.141592653589793, device=x.device, dtype=x.dtype)))) * torch.exp(
        -0.5 * ((x - mean) * inv) ** 2
    )

## Redistribute segmentation values transform (GPU)
class RandomRedistributeSegGPU(ImageOnlyTransform):
    """Redistribute image values using segmentation regions (GPU version).

    Mirrors the CPU `RedistributeTransform` behavior using GPU-friendly ops.
    Works with inputs shaped [N, C, H, W] or [N, C, D, H, W].
    """

    def __init__(
        self,
        in_seg: float = 0.2,
        apply_to_channel: list[int] = [0],
        retain_stats: bool = False,
        same_on_batch: bool = False,
        p: float = 1.0,
        keepdim: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, keepdim=keepdim)
        self.in_seg = in_seg
        self.apply_to_channel = apply_to_channel
        self.retain_stats = retain_stats

    @torch.no_grad()
    def apply_transform(
        self, input: Tensor, params: Dict[str, Tensor], flags: Dict[str, Any], transform: Optional[Tensor] = None
    ) -> Tensor:
        # Expect segmentation provided in params: shape [N, 1, ...] or [N, C_seg, ...]
        if 'seg' not in params:
            return input
        seg = params['seg']
        if seg.dim() != input.dim():
            # Allow seg [N, ...] by adding channel dim
            if seg.dim() == input.dim() - 1:
                seg = seg.unsqueeze(1)
            else:
                return input

        spatial_dims = input.dim() - 2
        if spatial_dims not in (2, 3):
            return input

        N = input.shape[0]

        # Apply per selected image channel and per batch sample
        for c in self.apply_to_channel:
            img_batch = input[:, c]  # (N, [...])
            # Sanitize incoming values to prevent NaN/Inf propagation
            img_batch = torch.nan_to_num(img_batch, nan=0.0, posinf=0.0, neginf=0.0)

            # Optionally retain original stats (vectorized per sample)
            if self.retain_stats:
                flat = img_batch.view(N, -1)
                orig_mean = flat.mean(dim=1)
                # Use unbiased=False to avoid NaNs for tiny tensors
                orig_std = flat.std(dim=1, unbiased=False)

            # Normalize entire batch to [0,1] per sample
            img_min = img_batch.view(N, -1).min(dim=1)[0].view(N, *([1] * (img_batch.dim()-1)))
            img_max = img_batch.view(N, -1).max(dim=1)[0].view(N, *([1] * (img_batch.dim()-1)))
            denom = (img_max - img_min).clamp_min(1e-6)
            x_batch = (img_batch - img_min) / denom

            # Iterate per sample (seg can differ in shape or labels per sample)
            for b in range(N):
                x = x_batch[b]
                seg_b = seg[b]  # shape (R, ...)

                # Quick skip if no foreground
                if (seg_b > 0).sum() == 0:
                    input[b, c] = x
                    continue

                # Decide redistribution mode once per sample
                # Scalar random flag for redistribution mode
                in_seg_bool = torch.rand((), device=input.device) <= self.in_seg

                # Binary masks for regions
                masks = seg_b.bool()  # (R, ...)
                R = masks.shape[0]

                # Vectorized dilation for all regions (3 iterations)
                dilated = masks.float()
                for _ in range(3):
                    if spatial_dims == 3:
                        dilated = F.max_pool3d(dilated.unsqueeze(0), 3, 1, 1).squeeze(0)
                    else:
                        dilated = F.max_pool2d(dilated.unsqueeze(0), 3, 1, 1).squeeze(0)
                dilated_excl = (dilated > 0) & (~masks)

                # Flatten for stats
                x_flat = x.view(1, -1)  # (1, S)
                mask_flat = masks.view(R, -1)
                dil_flat = dilated_excl.view(R, -1)

                # Region counts
                counts = mask_flat.sum(dim=1).clamp_min(1)
                # Means
                means = (mask_flat * x_flat).sum(dim=1) / counts
                # Std (compute variance then sqrt) avoid indexing overhead
                diffs = (x_flat - means.view(R,1)) * mask_flat
                vars = (diffs * diffs).sum(dim=1) / counts.clamp_min(1)
                stds = vars.sqrt()

                # Dilated stats
                dil_counts = dil_flat.sum(dim=1).clamp_min(1)
                dil_means = (dil_flat * x_flat).sum(dim=1) / dil_counts
                dil_diffs = (x_flat - dil_means.view(R,1)) * dil_flat
                dil_vars = (dil_diffs * dil_diffs).sum(dim=1) / dil_counts
                dil_stds = dil_vars.sqrt()

                # redist_std per region
                redist_std = torch.maximum(
                    torch.rand(R, device=input.device) * 0.2 + 0.4 * torch.abs((means - dil_means) * stds / (dil_stds + 1e-6)),
                    torch.full((R,), 0.01, device=input.device, dtype=input.dtype)
                )

                # Build additive term
                to_add = torch.zeros_like(x)
                rand_sign = (2 * torch.rand(R, device=input.device) - 1)  # random sign factor per region
                if in_seg_bool.item():
                    # Only inside region
                    for r in range(R):
                        if counts[r] == 0:  # skip empty
                            continue
                        region_vals = x[mask_flat[r].view(x.shape)]
                        pdf_vals = _normal_pdf(region_vals, means[r], redist_std[r]) * rand_sign[r]
                        to_add[mask_flat[r].view(x.shape)] += pdf_vals
                else:
                    # Global additive influence per region
                    pdf_all = []
                    for r in range(R):
                        if counts[r] == 0:
                            continue
                        pdf_all.append(_normal_pdf(x, means[r], redist_std[r]) * rand_sign[r])
                    if pdf_all:
                        to_add += torch.stack(pdf_all, dim=0).sum(dim=0)

                # Normalize to_add if non-zero
                tmin, tmax = to_add.min(), to_add.max()
                if (tmax - tmin) > 1e-8:
                    x = x + 2 * (to_add - tmin) / (tmax - tmin + 1e-6)

                # Restore stats
                if self.retain_stats:
                    mean = x.mean()
                    # Use unbiased=False to avoid NaNs on degenerate shapes
                    std = x.std(unbiased=False)
                    x = (x - mean) / torch.clamp(std, min=1e-7)
                    x = x * orig_std[b] + orig_mean[b]

                # Final safety: check if nan/inf appeared
                if torch.isnan(x).any() or torch.isinf(x).any():
                    print(f"Warning nan: {self.__class__.__name__}", flush=True)
                    continue
                input[b, c] = x

        return input



import os, json

import torch.nn as nn
import torch
import numpy as np

from auglab.transforms.gpu.contrast import RandomConvTransformGPU, RandomGaussianNoiseGPU, RandomBrightnessGPU, RandomGammaGPU, RandomFunctionGPU, \
RandomHistogramEqualizationGPU, RandomInverseGPU, RandomBiasFieldGPU, RandomContrastGPU, ZscoreNormalizationGPU, RandomClampGPU
from auglab.transforms.gpu.spatial import RandomAffine3DCustom, RandomLowResTransformGPU, RandomFlipTransformGPU, RandomAcqTransformGPU
from auglab.transforms.gpu.fromSeg import RandomRedistributeSegGPU
from auglab.transforms.gpu.base import AugmentationSequentialCustom

class AugTransformsGPU(AugmentationSequentialCustom):
    """
    Module to perform data augmentation on GPU.
    """
    def __init__(self, json_path: str):
        # Load transform parameters from JSON
        config_path = os.path.join(json_path)
        with open(config_path, 'r') as f:
            config = json.load(f)
        
        if 'GPU' in config.keys():
            self.transform_params = config['GPU']
        else:
            self.transform_params = config

        transforms = self._build_transforms()
        super().__init__(*transforms, data_keys=["input", "mask"], same_on_batch=True) # Same_on_batch to ensure mask are aligned with images correctly (custom) see AugmentationSequentialOpsCustom in base.py

    def _build_transforms(self) -> list[nn.Module]:
        transforms = []

        # Clamping transform
        clamp_params = self.transform_params.get('ClampTransform')
        if clamp_params is not None:
            transforms.append(RandomClampGPU(
                max_clamp_amount=clamp_params.get('max_clamp_amount', 0.0),
                in_seg=clamp_params.get('in_seg', 0.0),
                out_seg=clamp_params.get('out_seg', 0.0),
                mix_in_out=clamp_params.get('mix_in_out', False),
                retain_stats=clamp_params.get('retain_stats', False),
                p=clamp_params.get('probability', 0),
            ))

        # Noise transforms
        noise_params = self.transform_params.get('GaussianNoiseTransform')
        if noise_params is not None:
            transforms.append(RandomGaussianNoiseGPU(
                mean=noise_params.get('mean', 0.0),
                std=noise_params.get('std', 1.0),
                in_seg=noise_params.get('in_seg', 0.0),
                out_seg=noise_params.get('out_seg', 0.0),
                mix_in_out=noise_params.get('mix_in_out', False),
                p=noise_params.get('probability', 0),
            ))
        
        # Gaussian blur
        gaussianblur_params = self.transform_params.get('GaussianBlurTransform')
        if gaussianblur_params is not None:
            transforms.append(RandomConvTransformGPU(
                kernel_type=gaussianblur_params.get('kernel_type', 'GaussianBlur'),
                in_seg=gaussianblur_params.get('in_seg', 0.0),
                out_seg=gaussianblur_params.get('out_seg', 0.0),
                mix_in_out=gaussianblur_params.get('mix_in_out', False),
                p=gaussianblur_params.get('probability', 0),
                sigma=gaussianblur_params.get('sigma', 1.0),
            ))

        # Brightness transforms
        brightness_params = self.transform_params.get('BrightnessTransform')
        if brightness_params is not None:
            transforms.append(RandomBrightnessGPU(
                brightness_range=brightness_params.get('brightness_range', [0.5, 1.5]),
                in_seg=brightness_params.get('in_seg', 0.0),
                out_seg=brightness_params.get('out_seg', 0.0),
                mix_in_out=brightness_params.get('mix_in_out', False),
                p=brightness_params.get('probability', 0),
            ))

        # Gamma transforms
        gamma_params = self.transform_params.get('GammaTransform')
        if gamma_params is not None:
            transforms.append(RandomGammaGPU(
                gamma_range=gamma_params.get('gamma_range', [0.7, 1.5]),
                p=gamma_params.get('probability', 0),
                invert_image=False,
                in_seg=gamma_params.get('in_seg', 0.0),
                out_seg=gamma_params.get('out_seg', 0.0),
                mix_in_out=gamma_params.get('mix_in_out', False),
                retain_stats=gamma_params.get('retain_stats', False),
            ))

        inv_gamma_params = self.transform_params.get('InvGammaTransform')
        if inv_gamma_params is not None:
            transforms.append(RandomGammaGPU(
                gamma_range=inv_gamma_params.get('gamma_range', [0.7, 1.5]),
                p=inv_gamma_params.get('probability', 0),
                in_seg=inv_gamma_params.get('in_seg', 0.0),
                out_seg=inv_gamma_params.get('out_seg', 0.0),
                mix_in_out=inv_gamma_params.get('mix_in_out', False),
                invert_image=True,
                retain_stats=inv_gamma_params.get('retain_stats', False),
            ))
        
        # nnUNetV2 Contrast transforms
        contrast_params = self.transform_params.get('ContrastTransform')
        if contrast_params is not None:
            transforms.append(RandomContrastGPU(
                contrast_range=contrast_params.get('contrast_range', [0.75, 1.25]),
                p=contrast_params.get('probability', 0),
                in_seg=contrast_params.get('in_seg', 0.0),
                out_seg=contrast_params.get('out_seg', 0.0),
                mix_in_out=contrast_params.get('mix_in_out', False),
                retain_stats=contrast_params.get('retain_stats', False)
            ))

        # Apply functions
        func_list = [
            lambda x: torch.log(1 + x),
            torch.sqrt,
            torch.sin,
            torch.exp,
            lambda x: 1/(1 + torch.exp(-x)),
        ]
        function_params = self.transform_params.get('FunctionTransform')
        if function_params is not None:
            for func in func_list:
                transforms.append(RandomFunctionGPU(
                    func=func,
                    p=function_params.get('probability', 0),
                    in_seg=function_params.get('in_seg', 0.0),
                    out_seg=function_params.get('out_seg', 0.0),
                    mix_in_out=function_params.get('mix_in_out', False),
                    retain_stats=function_params.get('retain_stats', False),
            ))
        
        # Inverse transform (max - pixel_value)
        inverse_params = self.transform_params.get('InverseTransform')
        if inverse_params is not None:
            transforms.append(RandomInverseGPU(
                p=inverse_params.get('probability', 0),
                in_seg=inverse_params.get('in_seg', 0.0),
                out_seg=inverse_params.get('out_seg', 0.0),
                mix_in_out=inverse_params.get('mix_in_out', False),
                retain_stats=inverse_params.get('retain_stats', False),
            ))
        
        # Histogram manipulations
        histo_params = self.transform_params.get('HistogramEqualizationTransform')
        if histo_params is not None:
            transforms.append(RandomHistogramEqualizationGPU(
                p=histo_params.get('probability', 0),
                in_seg=histo_params.get('in_seg', 0.0),
                out_seg=histo_params.get('out_seg', 0.0),
                mix_in_out=histo_params.get('mix_in_out', False),
                retain_stats=histo_params.get('retain_stats', False),
            ))

        # Shape transforms (Cropping and Simulating low resolution)
        lowres_params = self.transform_params.get('SimulateLowResTransform')
        if lowres_params is not None:
            transforms.append(RandomLowResTransformGPU(
                p=lowres_params.get('probability', 0),
                scale=lowres_params.get('scale', [0.3, 1.0]),
                crop=lowres_params.get('crop', [1.0, 1.0]),
                same_on_batch=lowres_params.get('same_on_batch', False)
        ))

        acq_params = self.transform_params.get('AcqTransform')
        if acq_params is not None:
            transforms.append(RandomAcqTransformGPU(
                p=acq_params.get('probability', 0),
                scale=acq_params.get('scale', [0.3, 1.0]),
                crop=acq_params.get('crop', [1.0, 1.0]),
                one_dim=True,
                same_on_batch=acq_params.get('same_on_batch', False)
        ))
        
        # Bias field artifact
        bias_field_params = self.transform_params.get('BiasFieldTransform')
        if bias_field_params is not None:
            transforms.append(RandomBiasFieldGPU(
                p=bias_field_params.get('probability', 0),
                in_seg=bias_field_params.get('in_seg', 0.0),
                out_seg=bias_field_params.get('out_seg', 0.0),
                mix_in_out=bias_field_params.get('mix_in_out', False),
                retain_stats=bias_field_params.get('retain_stats', False),
                coefficients=bias_field_params.get('coefficients', 0.5),
            ))
        
        # Redistribute segmentation values transform
        redistribute_params = self.transform_params.get('RedistributeSegTransform')
        if redistribute_params is not None:
            transforms.append(RandomRedistributeSegGPU(
                in_seg=redistribute_params.get('in_seg', 0.2),
                retain_stats=redistribute_params.get('retain_stats', False),
                p=redistribute_params.get('probability', 0),
            ))

        # Flipping transforms
        flip_params = self.transform_params.get('FlipTransform')
        if flip_params is not None:
            transforms.append(RandomFlipTransformGPU(
                flip_axis=flip_params.get('flip_axis', [0]),
                p=flip_params.get('probability', 0),
                same_on_batch=flip_params.get('same_on_batch', False),
                keepdim=flip_params.get('keepdim', True)
            ))

        # Artifacts generation (Not implemented on GPU yet)

        # Spatial transforms
        affine_params = self.transform_params.get('AffineTransform')
        if affine_params is not None:
            transforms.append(RandomAffine3DCustom(
                degrees=affine_params.get('degrees', 10),
                translate=affine_params.get('translate', [0.1, 0.1, 0.1]),
                scale=affine_params.get('scale', [0.9, 1.1]),
                shears=affine_params.get('shear', [-10, 10, -10, 10, -10, 10]),
                resample=affine_params.get('resample', "bilinear"),
                p=affine_params.get('probability', 0)
            ))

        # Elastic transforms (Not implemented on GPU yet)

        # Scharr filter
        scharr_params = self.transform_params.get('ScharrTransform')
        if scharr_params is not None:
            transforms.append(RandomConvTransformGPU(
                kernel_type=scharr_params.get('kernel_type', 'Scharr'),
                p=scharr_params.get('probability', 0),
                in_seg=scharr_params.get('in_seg', 0.0),
                out_seg=scharr_params.get('out_seg', 0.0),
                mix_in_out=scharr_params.get('mix_in_out', False),
                retain_stats=scharr_params.get('retain_stats', True),
                absolute=scharr_params.get('absolute', True),
            ))

        # Unsharp masking
        unsharp_params = self.transform_params.get('UnsharpMaskTransform')
        if unsharp_params is not None:
            transforms.append(RandomConvTransformGPU(
                kernel_type=unsharp_params.get('kernel_type', 'UnsharpMask'),
                p=unsharp_params.get('probability', 0),
                in_seg=unsharp_params.get('in_seg', 0.0),
                out_seg=unsharp_params.get('out_seg', 0.0),
                mix_in_out=unsharp_params.get('mix_in_out', False),
                sigma=unsharp_params.get('sigma', 1.0),
                unsharp_amount=unsharp_params.get('unsharp_amount', 1.5),
        ))
            
        # RandomConv transform
        randconv_params = self.transform_params.get('RandomConvTransform')
        if randconv_params is not None:
            transforms.append(RandomConvTransformGPU(
                kernel_type=randconv_params.get('kernel_type', 'RandConv'),
                p=randconv_params.get('probability', 0),
                in_seg=randconv_params.get('in_seg', 0.0),
                out_seg=randconv_params.get('out_seg', 0.0),
                mix_in_out=randconv_params.get('mix_in_out', False),
                retain_stats=randconv_params.get('retain_stats', False),
                kernel_sizes=randconv_params.get('kernel_sizes', [1,3,5,7]),
                mix_prob=randconv_params.get('mix_prob', 0.0),
        ))
        
        # Z-score normalization
        zscore_params = self.transform_params.get('ZscoreNormalizationTransform')
        if zscore_params is not None:
            transforms.append(ZscoreNormalizationGPU(
                p=zscore_params.get('probability', 0)
            ))

        return transforms

def normalize(arr: np.ndarray) -> np.ndarray:
    """
    Normalize a tensor to the range [0, 1].
    """
    min_val = np.min(arr)
    max_val = np.max(arr)
    normalized_arr = (arr - min_val) / (max_val - min_val + 1e-8)
    return normalized_arr

def pad_numpy_array(arr, shape):
    """
    Pad a numpy array to the desired shape with zeros.
    """
    # Calculate padding needed for each dimension
    pad_width = [(max(0, shape[i] - arr.shape[i]) // 2, max(0, shape[i] - arr.shape[i]) - max(0, shape[i] - arr.shape[i]) // 2) for i in range(len(shape))]
    padded_arr = np.pad(arr, pad_width, mode='constant', constant_values=0)
    return padded_arr

if __name__ == "__main__":
    # Example usage
    import importlib
    import auglab.configs as configs
    from auglab.utils.image import Image, resample_nib

    configs_path = importlib.resources.files(configs)
    json_path = configs_path / "transform_params_gpu.json"
    augmentor = AugTransformsGPU(json_path)

    # Load images and masks tensors
    img_path = '/home/GRAMES.POLYMTL.CA/p118739/data_nvme_p118739/data/datasets/data-multi-subject/sub-amu02/anat/sub-amu02_T1w.nii.gz'
    img = Image(img_path).change_orientation('RSP')
    img = resample_nib(img, new_size=[1,1,1], new_size_type='mm', interpolation='linear')
    img_tensor = torch.from_numpy(img.data.copy()).to(torch.float32)

    seg_path = '/home/GRAMES.POLYMTL.CA/p118739/data_nvme_p118739/data/datasets/data-multi-subject/derivatives/labels/sub-amu02/anat/sub-amu02_T1w_label-spine_dseg.nii.gz'
    seg = Image(seg_path).change_orientation('RSP')
    seg = resample_nib(seg, new_size=[1,1,1], new_size_type='mm', interpolation='nn')
    seg_tensor_all = torch.from_numpy(seg.data.copy())
    
    img2_path = '/home/GRAMES.POLYMTL.CA/p118739/data_nvme_p118739/data/datasets/spider-challenge-2023/sub-002/anat/sub-002_acq-lowresSag_T2w.nii.gz'
    img2 = Image(img2_path).change_orientation('RSP')
    img2 = resample_nib(img2, new_size=[1,1,1], new_size_type='mm', interpolation='linear')
    img2_tensor = torch.from_numpy(img2.data.copy()).to(torch.float32)

    seg2_path = '/home/GRAMES.POLYMTL.CA/p118739/data_nvme_p118739/data/datasets/spider-challenge-2023/derivatives/labels/sub-002/anat/sub-002_acq-lowresSag_T2w_label-spine_dseg.nii.gz'
    seg2 = Image(seg2_path).change_orientation('RSP')
    seg2 = resample_nib(seg2, new_size=[1,1,1], new_size_type='mm', interpolation='nn')
    seg2_tensor_all = torch.from_numpy(seg2.data.copy())

    # Combine two images to same size
    new_shape = []
    for dim in range(3):
        size1 = img_tensor.shape[dim]
        size2 = img2_tensor.shape[dim]
        min_size = min(size1, size2)
        new_shape.append(min_size)
    
    new_img_tensor = torch.zeros(new_shape)
    new_img2_tensor = torch.zeros(new_shape)
    new_seg_tensor_all = torch.zeros(new_shape)
    new_seg2_tensor_all = torch.zeros(new_shape)

    gap = (torch.tensor(img_tensor.shape) - torch.tensor(new_shape)) // 2
    gap2 = (torch.tensor(img2_tensor.shape) - torch.tensor(new_shape)) // 2
    new_img_tensor = img_tensor[gap[0]:gap[0]+new_shape[0], gap[1]:gap[1]+new_shape[1], gap[2]:gap[2]+new_shape[2]]
    new_img2_tensor = img2_tensor[gap2[0]:gap2[0]+new_shape[0], gap2[1]:gap2[1]+new_shape[1], gap2[2]:gap2[2]+new_shape[2]]
    new_seg_tensor_all = seg_tensor_all[gap[0]:gap[0]+new_shape[0], gap[1]:gap[1]+new_shape[1], gap[2]:gap[2]+new_shape[2]]
    new_seg2_tensor_all = seg2_tensor_all[gap2[0]:gap2[0]+new_shape[0], gap2[1]:gap2[1]+new_shape[1], gap2[2]:gap2[2]+new_shape[2]]

    # Add segmentation values to different channels
    seg_tensor = torch.zeros((1, 5, *new_seg_tensor_all.shape))
    for i, value in enumerate([12, 13, 14, 15, 16]):
        seg_tensor[0, i] = (new_seg_tensor_all == value)
    
    seg2_tensor = torch.zeros((1, 5, *new_seg2_tensor_all.shape))
    for i, value in enumerate([50, 45, 44, 43, 42]):
        seg2_tensor[0, i] = (new_seg2_tensor_all == value)

    # Format tensors to match expected input shape (B, C, D, H, W)
    img_tensor = torch.cat([new_img_tensor.unsqueeze(0), new_seg_tensor_all.bool().int().unsqueeze(0)], dim=0).unsqueeze(0)  # Add batch dimension and second channel
    img2_tensor = torch.cat([new_img2_tensor.unsqueeze(0), new_seg2_tensor_all.bool().int().unsqueeze(0)], dim=0).unsqueeze(0)  # Add batch dimension and second channel

    # Add batch 
    img_tensor = torch.cat([img_tensor, img2_tensor], dim=0)
    seg_tensor = torch.cat([seg_tensor, seg2_tensor], dim=0)

    # Move to GPU
    img_tensor = img_tensor.cuda()
    seg_tensor = seg_tensor.cuda()
    augmentor = augmentor.cuda()

    # Apply augmentations
    augmented_img, augmented_seg = augmentor(img_tensor.clone(), seg_tensor.clone())

    if augmented_img.shape != img_tensor.shape:
        raise ValueError("Augmented image shape does not match input shape.")
    if augmented_seg.shape != seg_tensor.shape:
        raise ValueError("Augmented segmentation shape does not match input shape.")
    # Check if nans are present
    if torch.isnan(augmented_img).any():
        raise ValueError("NaNs found in augmented image.")
    if torch.isnan(augmented_seg).any():
        raise ValueError("NaNs found in augmented segmentation.")
    
    import cv2
    import numpy as np
    import warnings, sys, os
    warnings.simplefilter("always")

    # Convert tensors to numpy arrays
    img_tensor_np = img_tensor.cpu().detach().numpy()
    seg_tensor_np = seg_tensor.cpu().detach().numpy()
    augmented_img_np = augmented_img.cpu().detach().numpy()
    augmented_seg_np = augmented_seg.cpu().detach().numpy()

    # Concatenate segmentation channels for visualization
    seg_tensor_np = np.sum(seg_tensor_np, axis=1)
    augmented_seg_np = np.sum(augmented_seg_np, axis=1)

    pad_shape = 2*(np.max(img_tensor_np.shape[2:]),)

    # Combine tensors into single output for visualization
    os.makedirs('img', exist_ok=True)
    img_line = np.concatenate([normalize(pad_numpy_array(img_tensor_np[0, 0, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(img_tensor_np[0, 0, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(img_tensor_np[0, 0, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    augmented_img_line = np.concatenate([normalize(pad_numpy_array(augmented_img_np[0, 0, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[0, 0, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[0, 0, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    seg_line = np.concatenate([normalize(pad_numpy_array(seg_tensor_np[0, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(seg_tensor_np[0, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(seg_tensor_np[0, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    augmented_seg_line = np.concatenate([normalize(pad_numpy_array(augmented_seg_np[0, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(augmented_seg_np[0, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(augmented_seg_np[0, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    not_augmented_channel_line = np.concatenate([normalize(pad_numpy_array(augmented_img_np[0, 1, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[0, 1, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[0, 1, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    combined_img = np.concatenate([img_line, seg_line, augmented_img_line, augmented_seg_line, not_augmented_channel_line], axis=0)
    cv2.imwrite('img/combined.png', combined_img*255)

    img_line2 = np.concatenate([normalize(pad_numpy_array(img_tensor_np[1, 0, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(img_tensor_np[1, 0, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(img_tensor_np[1, 0, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    augmented_img_line2 = np.concatenate([normalize(pad_numpy_array(augmented_img_np[1, 0, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[1, 0, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[1, 0, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    seg_line2 = np.concatenate([normalize(pad_numpy_array(seg_tensor_np[1, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(seg_tensor_np[1, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(seg_tensor_np[1, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    augmented_seg_line2 = np.concatenate([normalize(pad_numpy_array(augmented_seg_np[1, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(augmented_seg_np[1, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(augmented_seg_np[1, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    not_augmented_channel_line2 = np.concatenate([normalize(pad_numpy_array(augmented_img_np[1, 1, img_tensor_np.shape[2] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[1, 1, :, :, img_tensor_np.shape[4] // 2], pad_shape)), normalize(pad_numpy_array(augmented_img_np[1, 1, :, img_tensor_np.shape[3] // 2, :], pad_shape))], axis=1)
    combined_img2 = np.concatenate([img_line2, seg_line2, augmented_img_line2, augmented_seg_line2, not_augmented_channel_line2], axis=0)
    cv2.imwrite('img/combined2.png', combined_img2*255)

    print(augmentor)


from nnunetv2.training.nnUNetTrainer.nnUNetTrainer import nnUNetTrainer
from nnunetv2.training.dataloading.data_loader import nnUNetDataLoader
from nnunetv2.utilities.default_n_proc_DA import get_allowed_n_proc_DA
from nnunetv2.training.dataloading.nnunet_dataset import infer_dataset_class
from nnunetv2.utilities.label_handling.label_handling import determine_num_input_channels
from torch.nn.parallel import DistributedDataParallel as DDP

from batchgenerators.dataloading.single_threaded_augmenter import SingleThreadedAugmenter
from batchgenerators.dataloading.nondet_multi_threaded_augmenter import NonDetMultiThreadedAugmenter
from typing import Tuple, Union, List
import numpy as np
from batchgeneratorsv2.helpers.scalar_type import RandomScalar
from batchgeneratorsv2.transforms.base.basic_transform import BasicTransform
from batchgeneratorsv2.transforms.nnunet.random_binary_operator import ApplyRandomBinaryOperatorTransform
from batchgeneratorsv2.transforms.nnunet.remove_connected_components import RemoveRandomConnectedComponentFromOneHotEncodingTransform
from batchgeneratorsv2.transforms.nnunet.seg_to_onehot import MoveSegAsOneHotToDataTransform
from batchgeneratorsv2.transforms.utils.compose import ComposeTransforms
from batchgeneratorsv2.transforms.utils.deep_supervision_downsampling import DownsampleSegForDSTransform
from batchgeneratorsv2.transforms.utils.nnunet_masking import MaskImageTransform
from batchgeneratorsv2.transforms.utils.random import RandomTransform
from batchgeneratorsv2.transforms.utils.remove_label import RemoveLabelTansform
from batchgeneratorsv2.transforms.utils.seg_to_regions import ConvertSegmentationToRegionsTransform
from batchgeneratorsv2.transforms.utils.pseudo2d import Convert3DTo2DTransform, Convert2DTo3DTransform
from batchgeneratorsv2.transforms.spatial.spatial import SpatialTransform

import os
import torch
import importlib
from torch import autocast
import json
import shutil

from nnunetv2.training.nnUNetTrainer.nnUNetTrainer import nnUNetTrainer
from nnunetv2.utilities.helpers import dummy_context

from auglab.transforms.cpu.transforms import AugTransforms
from auglab.transforms.cpu.contrast import ZscoreNormalization
import auglab.configs as configs
from auglab.transforms.gpu.transforms import AugTransformsGPU
from auglab.trainers.utils import DownsampleSegForDSTransformCustom

class nnUNetTrainerDAExt(nnUNetTrainer):
    def __init__(self, plans: dict, configuration: str, fold: int, dataset_json: dict,
                 device: torch.device = torch.device('cuda')):
        super().__init__(plans, configuration, fold, dataset_json, device)
    
    def configure_rotation_dummyDA_mirroring_and_inital_patch_size(self):
        rotation_for_DA, do_dummy_2d_data_aug, initial_patch_size, mirror_axes = \
            super().configure_rotation_dummyDA_mirroring_and_inital_patch_size()
        # Remove mirroring
        mirror_axes = None
        self.inference_allowed_mirroring_axes = None
        return rotation_for_DA, do_dummy_2d_data_aug, initial_patch_size, mirror_axes
    
    @staticmethod
    def get_training_transforms(
            patch_size: Union[np.ndarray, Tuple[int]],
            rotation_for_DA: RandomScalar,
            deep_supervision_scales: Union[List, Tuple, None],
            mirror_axes: Tuple[int, ...],
            do_dummy_2d_data_aug: bool,
            use_mask_for_norm: List[bool] = None,
            is_cascaded: bool = False,
            foreground_labels: Union[Tuple[int, ...], List[int]] = None,
            regions: List[Union[List[int], Tuple[int, ...], int]] = None,
            ignore_label: int = None,
            retain_stats: bool = False
    ) -> BasicTransform:
        transforms = []

        ### Adds transforms
        # Load transform parameters from json file
        configs_path = importlib.resources.files(configs)
        json_path = os.environ.get("AUGLAB_PARAMS_CPU_JSON", str(configs_path / "transform_params.json"))
        transforms.append(AugTransforms(json_path=json_path))

        if do_dummy_2d_data_aug:
            ignore_axes = (0,)
            transforms.append(Convert3DTo2DTransform())
            patch_size_spatial = patch_size[1:]
        else:
            patch_size_spatial = patch_size
            ignore_axes = None
        transforms.append(
            SpatialTransform(
                patch_size_spatial, 
                patch_center_dist_from_border=0, 
                random_crop=False, 
                p_elastic_deform=0,
                p_rotation=0,
                rotation=rotation_for_DA, 
                p_scaling=0, 
                scaling=(0.7, 1.4), 
                p_synchronize_scaling_across_axes=1,
                bg_style_seg_sampling=False, 
                mode_seg='nearest'
            )
        )

        if do_dummy_2d_data_aug:
            transforms.append(Convert2DTo3DTransform())
        
        if use_mask_for_norm is not None and any(use_mask_for_norm):
            transforms.append(MaskImageTransform(
                apply_to_channels=[i for i in range(len(use_mask_for_norm)) if use_mask_for_norm[i]],
                channel_idx_in_seg=0,
                set_outside_to=0,
            ))

        transforms.append(
            RemoveLabelTansform(-1, 0)
        )

        # The following augmentations are related to special nnunet executions
        if is_cascaded:
            assert foreground_labels is not None, 'We need foreground_labels for cascade augmentations'
            transforms.append(
                MoveSegAsOneHotToDataTransform(
                    source_channel_idx=1,
                    all_labels=foreground_labels,
                    remove_channel_from_source=True
                )
            )
            transforms.append(
                RandomTransform(
                    ApplyRandomBinaryOperatorTransform(
                        channel_idx=list(range(-len(foreground_labels), 0)),
                        strel_size=(1, 8),
                        p_per_label=1
                    ), apply_probability=0.4
                )
            )
            transforms.append(
                RandomTransform(
                    RemoveRandomConnectedComponentFromOneHotEncodingTransform(
                        channel_idx=list(range(-len(foreground_labels), 0)),
                        fill_with_other_class_p=0,
                        dont_do_if_covers_more_than_x_percent=0.15,
                        p_per_label=1
                    ), apply_probability=0.2
                )
            )

        if regions is not None:
            # the ignore label must also be converted
            transforms.append(
                ConvertSegmentationToRegionsTransform(
                    regions=list(regions) + [ignore_label] if ignore_label is not None else regions,
                    channel_in_seg=0
                )
            )

        if deep_supervision_scales is not None:
            transforms.append(DownsampleSegForDSTransform(ds_scales=deep_supervision_scales))

        return ComposeTransforms(transforms)


class nnUNetTrainerDAExtGPU(nnUNetTrainer):
    def __init__(self, plans: dict, configuration: str, fold: int, dataset_json: dict,
                 device: torch.device = torch.device('cuda')):
        super().__init__(plans, configuration, fold, dataset_json, device)

        # Load transform parameters from json file
        configs_path = importlib.resources.files(configs)
        json_path = os.environ.get("AUGLAB_PARAMS_GPU_JSON", str(configs_path / "transform_params_gpu.json"))
        self.transforms = AugTransformsGPU(json_path=json_path).to(self.device)
        print(f'Using AugLab GPU transforms with parameters from: {json_path}')

        # Copy json transfrom parameters to output folder
        shutil.copy(
            json_path,
            os.path.join(self.output_folder, 'transform_params_gpu_used_for_training.json')
        )

    def configure_rotation_dummyDA_mirroring_and_inital_patch_size(self):
        rotation_for_DA, do_dummy_2d_data_aug, initial_patch_size, mirror_axes = \
            super().configure_rotation_dummyDA_mirroring_and_inital_patch_size()
        # Remove mirroring
        mirror_axes = None
        self.inference_allowed_mirroring_axes = None
        return rotation_for_DA, do_dummy_2d_data_aug, initial_patch_size, mirror_axes

    @staticmethod
    def get_training_transforms(
            patch_size: Union[np.ndarray, Tuple[int]],
            rotation_for_DA: RandomScalar,
            deep_supervision_scales: Union[List, Tuple, None],
            mirror_axes: Tuple[int, ...],
            do_dummy_2d_data_aug: bool,
            use_mask_for_norm: List[bool] = None,
            is_cascaded: bool = False,
            foreground_labels: Union[Tuple[int, ...], List[int]] = None,
            regions: List[Union[List[int], Tuple[int, ...], int]] = None,
            ignore_label: int = None,
            retain_stats: bool = False
    ) -> BasicTransform:
        transforms = []

        ### Keep some nnunet transforms
        if do_dummy_2d_data_aug:
            ignore_axes = (0,)
            transforms.append(Convert3DTo2DTransform())
            patch_size_spatial = patch_size[1:]
        else:
            patch_size_spatial = patch_size
            ignore_axes = None
        transforms.append(
            SpatialTransform(
                patch_size_spatial, 
                patch_center_dist_from_border=0, 
                random_crop=False, 
                p_elastic_deform=0,
                p_rotation=0,
                rotation=rotation_for_DA, 
                p_scaling=0, 
                scaling=(0.7, 1.4), 
                p_synchronize_scaling_across_axes=1,
                bg_style_seg_sampling=False, 
                mode_seg='nearest'
            )
        )

        if do_dummy_2d_data_aug:
            transforms.append(Convert2DTo3DTransform())
        
        if use_mask_for_norm is not None and any(use_mask_for_norm):
            transforms.append(MaskImageTransform(
                apply_to_channels=[i for i in range(len(use_mask_for_norm)) if use_mask_for_norm[i]],
                channel_idx_in_seg=0,
                set_outside_to=0,
            ))

        transforms.append(
            RemoveLabelTansform(-1, 0)
        )

        # The following augmentations are related to special nnunet executions
        if is_cascaded:
            assert foreground_labels is not None, 'We need foreground_labels for cascade augmentations'
            transforms.append(
                MoveSegAsOneHotToDataTransform(
                    source_channel_idx=1,
                    all_labels=foreground_labels,
                    remove_channel_from_source=True
                )
            )
            transforms.append(
                RandomTransform(
                    ApplyRandomBinaryOperatorTransform(
                        channel_idx=list(range(-len(foreground_labels), 0)),
                        strel_size=(1, 8),
                        p_per_label=1
                    ), apply_probability=0.4
                )
            )
            transforms.append(
                RandomTransform(
                    RemoveRandomConnectedComponentFromOneHotEncodingTransform(
                        channel_idx=list(range(-len(foreground_labels), 0)),
                        fill_with_other_class_p=0,
                        dont_do_if_covers_more_than_x_percent=0.15,
                        p_per_label=1
                    ), apply_probability=0.2
                )
            )

        if regions is not None:
            # the ignore label must also be converted
            transforms.append(
                ConvertSegmentationToRegionsTransform(
                    regions=list(regions) + [ignore_label] if ignore_label is not None else regions,
                    channel_in_seg=0
                )
            )
        
        transforms.append(ZscoreNormalization())

        # NOTE: DownsampleSegForDSTransform is now handled in train_step for GPU augmentations
        # if deep_supervision_scales is not None:
        #     transforms.append(DownsampleSegForDSTransform(ds_scales=deep_supervision_scales))

        return ComposeTransforms(transforms)

    @staticmethod
    def get_validation_transforms(
            deep_supervision_scales: Union[List, Tuple, None],
            is_cascaded: bool = False,
            foreground_labels: Union[Tuple[int, ...], List[int]] = None,
            regions: List[Union[List[int], Tuple[int, ...], int]] = None,
            ignore_label: int = None,
    ) -> BasicTransform:
        transforms = []
        transforms.append(
            RemoveLabelTansform(-1, 0)
        )

        if is_cascaded:
            transforms.append(
                MoveSegAsOneHotToDataTransform(
                    source_channel_idx=1,
                    all_labels=foreground_labels,
                    remove_channel_from_source=True
                )
            )

        if regions is not None:
            # the ignore label must also be converted
            transforms.append(
                ConvertSegmentationToRegionsTransform(
                    regions=list(regions) + [ignore_label] if ignore_label is not None else regions,
                    channel_in_seg=0
                )
            )
        
        transforms.append(ZscoreNormalization())

        if deep_supervision_scales is not None:
            transforms.append(DownsampleSegForDSTransform(ds_scales=deep_supervision_scales))
        return ComposeTransforms(transforms)
    
    def train_step(self, batch: dict) -> dict:
        data = batch['data']
        target = batch['target']

        data = data.to(self.device, non_blocking=True)
        # Now target should be a single tensor, not a list
        target = target.to(self.device, non_blocking=True)
        # if isinstance(target, list):
        #     target = [i.to(self.device, non_blocking=True) for i in target]
        # else:
        #     target = target.to(self.device, non_blocking=True)

        self.optimizer.zero_grad(set_to_none=True)
        # Autocast can be annoying
        # If the device_type is 'cpu' then it's slow as heck and needs to be disabled.
        # If the device_type is 'mps' then it will complain that mps is not implemented, even if enabled=False is set. Whyyyyyyy. (this is why we don't make use of enabled=False)
        # So autocast will only be active if we have a cuda device.
        with autocast(self.device.type, enabled=True) if self.device.type == 'cuda' else dummy_context():            
            # Apply GPU augmentations to full-resolution data/target
            data, target = self.transforms(data, target)
    
            # Create multi-scale targets for deep supervision after augmentation
            deep_supervision_scales = self._get_deep_supervision_scales()
            if deep_supervision_scales is not None:
                ds_transform = DownsampleSegForDSTransformCustom(ds_scales=deep_supervision_scales)
                target = ds_transform(target)
    
            output = self.network(data)
            # del data
            l = self.loss(output, target)

        if self.grad_scaler is not None:
            self.grad_scaler.scale(l).backward()
            self.grad_scaler.unscale_(self.optimizer)
            torch.nn.utils.clip_grad_norm_(self.network.parameters(), 12)
            self.grad_scaler.step(self.optimizer)
            self.grad_scaler.update()
        else:
            l.backward()
            torch.nn.utils.clip_grad_norm_(self.network.parameters(), 12)
            self.optimizer.step()
        return {'loss': l.detach().cpu().numpy()}


class nnUNetTrainerDAExtHybrid(nnUNetTrainer):
    def __init__(self, plans: dict, configuration: str, fold: int, dataset_json: dict,
                 device: torch.device = torch.device('cuda')):
        super().__init__(plans, configuration, fold, dataset_json, device)

        # Load transform parameters from json file
        configs_path = importlib.resources.files(configs)
        json_path = os.environ.get("AUGLAB_PARAMS_HYBRID_JSON", str(configs_path / "transform_params_hybrid.json"))
        self.transforms = AugTransformsGPU(json_path=json_path).to(self.device)
        print(f'Using AugLab hybrid transforms with parameters from: {json_path}')

        # Copy json transfrom parameters to output folder
        shutil.copy(
            json_path,
            os.path.join(self.output_folder, 'transform_params_hybrid_used_for_training.json')
        )

    def configure_rotation_dummyDA_mirroring_and_inital_patch_size(self):
        rotation_for_DA, do_dummy_2d_data_aug, initial_patch_size, mirror_axes = \
            super().configure_rotation_dummyDA_mirroring_and_inital_patch_size()
        # Remove mirroring
        mirror_axes = None
        self.inference_allowed_mirroring_axes = None
        return rotation_for_DA, do_dummy_2d_data_aug, initial_patch_size, mirror_axes

    @staticmethod
    def get_training_transforms(
            patch_size: Union[np.ndarray, Tuple[int]],
            rotation_for_DA: RandomScalar,
            deep_supervision_scales: Union[List, Tuple, None],
            mirror_axes: Tuple[int, ...],
            do_dummy_2d_data_aug: bool,
            use_mask_for_norm: List[bool] = None,
            is_cascaded: bool = False,
            foreground_labels: Union[Tuple[int, ...], List[int]] = None,
            regions: List[Union[List[int], Tuple[int, ...], int]] = None,
            ignore_label: int = None,
            retain_stats: bool = False
    ) -> BasicTransform:
        transforms = []

        ### Adds transforms
        # Load transform parameters from json file
        configs_path = importlib.resources.files(configs)
        json_path = os.environ.get("AUGLAB_PARAMS_HYBRID_JSON", str(configs_path / "transform_params_hybrid.json"))
        transforms.append(AugTransforms(json_path=json_path, do_dummy_2d_data_aug=do_dummy_2d_data_aug, patch_size=patch_size, rotation_for_DA=rotation_for_DA, mirror_axes=mirror_axes))
        
        if use_mask_for_norm is not None and any(use_mask_for_norm):
            transforms.append(MaskImageTransform(
                apply_to_channels=[i for i in range(len(use_mask_for_norm)) if use_mask_for_norm[i]],
                channel_idx_in_seg=0,
                set_outside_to=0,
            ))

        transforms.append(
            RemoveLabelTansform(-1, 0)
        )

        # The following augmentations are related to special nnunet executions
        if is_cascaded:
            assert foreground_labels is not None, 'We need foreground_labels for cascade augmentations'
            transforms.append(
                MoveSegAsOneHotToDataTransform(
                    source_channel_idx=1,
                    all_labels=foreground_labels,
                    remove_channel_from_source=True
                )
            )
            transforms.append(
                RandomTransform(
                    ApplyRandomBinaryOperatorTransform(
                        channel_idx=list(range(-len(foreground_labels), 0)),
                        strel_size=(1, 8),
                        p_per_label=1
                    ), apply_probability=0.4
                )
            )
            transforms.append(
                RandomTransform(
                    RemoveRandomConnectedComponentFromOneHotEncodingTransform(
                        channel_idx=list(range(-len(foreground_labels), 0)),
                        fill_with_other_class_p=0,
                        dont_do_if_covers_more_than_x_percent=0.15,
                        p_per_label=1
                    ), apply_probability=0.2
                )
            )

        if regions is not None:
            # the ignore label must also be converted
            transforms.append(
                ConvertSegmentationToRegionsTransform(
                    regions=list(regions) + [ignore_label] if ignore_label is not None else regions,
                    channel_in_seg=0
                )
            )
        
        # NOTE: DownsampleSegForDSTransform is now handled in train_step for GPU augmentations
        # if deep_supervision_scales is not None:
        #     transforms.append(DownsampleSegForDSTransform(ds_scales=deep_supervision_scales))

        return ComposeTransforms(transforms)
    
    def train_step(self, batch: dict) -> dict:
        data = batch['data']
        target = batch['target']

        data = data.to(self.device, non_blocking=True)
        # Now target should be a single tensor, not a list
        target = target.to(self.device, non_blocking=True)
        # if isinstance(target, list):
        #     target = [i.to(self.device, non_blocking=True) for i in target]
        # else:
        #     target = target.to(self.device, non_blocking=True)

        self.optimizer.zero_grad(set_to_none=True)
        # Autocast can be annoying
        # If the device_type is 'cpu' then it's slow as heck and needs to be disabled.
        # If the device_type is 'mps' then it will complain that mps is not implemented, even if enabled=False is set. Whyyyyyyy. (this is why we don't make use of enabled=False)
        # So autocast will only be active if we have a cuda device.
        with autocast(self.device.type, enabled=True) if self.device.type == 'cuda' else dummy_context():            
            # Apply GPU augmentations to full-resolution data/target
            data, target = self.transforms(data, target)
    
            # Create multi-scale targets for deep supervision after augmentation
            deep_supervision_scales = self._get_deep_supervision_scales()
            if deep_supervision_scales is not None:
                ds_transform = DownsampleSegForDSTransformCustom(ds_scales=deep_supervision_scales)
                target = ds_transform(target)
    
            output = self.network(data)
            # del data
            l = self.loss(output, target)

        if self.grad_scaler is not None:
            self.grad_scaler.scale(l).backward()
            self.grad_scaler.unscale_(self.optimizer)
            torch.nn.utils.clip_grad_norm_(self.network.parameters(), 12)
            self.grad_scaler.step(self.optimizer)
            self.grad_scaler.update()
        else:
            l.backward()
            torch.nn.utils.clip_grad_norm_(self.network.parameters(), 12)
            self.optimizer.step()
        return {'loss': l.detach().cpu().numpy()}

